"""
Checks that Pylint does not complain about using Model and Manager methods
"""
#  pylint: disable=missing-docstring
from django.db import models


class SomeModel(models.Model):
    pass


if __name__ == '__main__':
    MODEL = SomeModel()
    MODEL.save()
    MODEL.delete()

    COUNT = SomeModel.objects.count()
    # added in django 1.6
    FIRST = SomeModel.objects.first()
    LAST = SomeModel.objects.last()



"""
Checks that Pylint does not complain about various
methods on Django model fields.
"""
#  pylint: disable=missing-docstring,wrong-import-position
from django.db import models
from django.db.models import ForeignKey, OneToOneField


class Author(models.Model):
    author_name = models.CharField(max_length=100)


class ISBN(models.Model):
    value = models.CharField(max_length=100)


class Book(models.Model):
    book_name = models.CharField(max_length=100)
    author = models.ForeignKey('Author', on_delete=models.CASCADE)
    isbn = models.OneToOneField(ISBN, on_delete=models.CASCADE)

    def get_isbn(self):
        return self.isbn.value

    def get_author_name(self):
        return self.author.author_name


class Fruit(models.Model):
    fruit_name = models.CharField(max_length=20)


class Seed(models.Model):
    fruit = ForeignKey(Fruit, on_delete=models.CASCADE)

    def get_fruit_name(self):
        return self.fruit.fruit_name


class User(models.Model):
    username = models.CharField(max_length=32)


class UserProfile(models.Model):
    user = OneToOneField(User, on_delete=models.CASCADE)

    def get_username(self):
        return self.user.username


class UserPreferences(models.Model):
    """
        Used for testing FK which refers to another model by
        string, not model class, see
        https://github.com/PyCQA/pylint-django/issues/35
    """
    user = ForeignKey('User', on_delete=models.CASCADE)



"""
This test ensures that a 'Meta' class defined on a Django model does
not raise warnings such as 'old-style-class' and 'too-few-public-methods'
"""
#  pylint: disable=missing-docstring

from django.db import models


class SomeModel(models.Model):
    class Meta:
        ordering = ('-id',)



"""
Checks that Pylint does not complain about duplicate
except blocks catching DoesNotExist exceptions:
https://github.com/PyCQA/pylint-django/issues/81
"""
#  pylint: disable=missing-docstring
from django.db import models


class Book(models.Model):
    name = models.CharField(max_length=100)


class Author(models.Model):
    name = models.CharField(max_length=100)


def dummy_func():
    try:
        print("foo")
    except Book.DoesNotExist:
        print("bar")
    except Author.DoesNotExist:
        print("baz")



from django.contrib.postgres import fields as django_fields
from psycopg2 import extras


# --------
# lists

class ArrayField(list, django_fields.ArrayField):
    pass


# --------
# dicts

class HStoreField(dict, django_fields.HStoreField):
    pass


class JSONField(dict, django_fields.JSONField):
    pass


# --------
# ranges

class RangeField(extras.Range, django_fields.RangeField):
    pass


class IntegerRangeField(extras.NumericRange, django_fields.IntegerRangeField):
    pass


class BigIntegerRangeField(extras.NumericRange, django_fields.BigIntegerRangeField):
    pass


class FloatRangeField(extras.NumericRange, django_fields.FloatRangeField):
    pass


class DateTimeRangeField(extras.DateTimeTZRange, django_fields.DateRangeField):
    pass


class DateRangeField(extras.DateRange, django_fields.DateRangeField):
    pass



"""
Checks that Pylint does not complain about foreign key sets on models
"""
#  pylint: disable=missing-docstring

from django.db import models


class SomeModel(models.Model):
    name = models.CharField(max_length=20)

    def get_others(self):
        return self.othermodel_set.all()

    def get_first(self):
        return self.othermodel_set.first()


class OtherModel(models.Model):
    count = models.IntegerField()
    something = models.ForeignKey(SomeModel, on_delete=models.CASCADE)


class ThirdModel(models.Model):
    whatever = models.ForeignKey(SomeModel, related_name='whatevs',
                                 on_delete=models.CASCADE)


def count_whatevers():
    if SomeModel().whatevs.exists():
        return SomeModel().whatevs.count()
    return -1



"""
Checks that Pylint does not complain about various
methods on Django form forms.
"""
#  pylint: disable=missing-docstring,R0904
from __future__ import print_function
from datetime import datetime, date
from django import forms


class ManyFieldsForm(forms.Form):

    booleanfield = forms.BooleanField()
    charfield = forms.CharField(max_length=40, null=True)
    datetimefield = forms.DateTimeField(auto_now_add=True)
    datefield = forms.DateField(auto_now_add=True)
    decimalfield = forms.DecimalField(max_digits=5, decimal_places=2)
    durationfield = forms.DurationField()
    emailfield = forms.EmailField()
    filefield = forms.FileField(name='test_file', upload_to='test')
    filepathfield = forms.FilePathField(path='/some/path')
    floatfield = forms.FloatField()
    genericipaddressfield = forms.GenericIPAddressField()
    imagefield = forms.ImageField(name='test_image', upload_to='test')
    intfield = forms.IntegerField(null=True)
    nullbooleanfield = forms.NullBooleanField()
    slugfield = forms.SlugField()
    timefield = forms.TimeField()
    urlfield = forms.URLField()

    def boolean_field_tests(self):
        print(self.booleanfield | True)
        print(self.nullbooleanfield | True)

    def string_field_tests(self):
        print(self.charfield.strip())
        print(self.charfield.upper())
        print(self.charfield.replace('x', 'y'))

        print(self.filepathfield.strip())
        print(self.filepathfield.upper())
        print(self.filepathfield.replace('x', 'y'))

        print(self.emailfield.strip())
        print(self.emailfield.upper())
        print(self.emailfield.replace('x', 'y'))

    def datetimefield_tests(self):
        now = datetime.now()
        print(now - self.datetimefield)
        print(self.datetimefield.ctime())

    def datefield_tests(self):
        now = date.today()
        print(now - self.datefield)
        print(self.datefield.isoformat())

    def decimalfield_tests(self):
        print(self.decimalfield.adjusted())

    def durationfield_tests(self):
        now = datetime.now()
        print(now - self.durationfield)
        print(self.durationfield.total_seconds())

    def filefield_tests(self):
        print(self.filefield)
        print(self.imagefield)

    def numberfield_tests(self):
        print(self.intfield + 5)



"""Models."""
from astroid import Const
from astroid.nodes import Assign
from astroid.nodes import ClassDef, FunctionDef, AssignName

from pylint.interfaces import IAstroidChecker
from pylint.checkers.utils import check_messages
from pylint.checkers import BaseChecker
from pylint.__pkginfo__ import numversion as PYLINT_VERSION

from pylint_django.__pkginfo__ import BASE_ID
from pylint_django.utils import node_is_subclass, PY3


MESSAGES = {
    'E%d01' % BASE_ID: ("__unicode__ on a model must be callable (%s)",
                        'model-unicode-not-callable',
                        "Django models require a callable __unicode__ method"),
    'W%d01' % BASE_ID: ("No __unicode__ method on model (%s)",
                        'model-missing-unicode',
                        "Django models should implement a __unicode__ method for string representation"),
    'W%d02' % BASE_ID: ("Found __unicode__ method on model (%s). Python3 uses __str__.",
                        'model-has-unicode',
                        "Django models should not implement a __unicode__ "
                        "method for string representation when using Python3"),
    'W%d03' % BASE_ID: ("Model does not explicitly define __unicode__ (%s)",
                        'model-no-explicit-unicode',
                        "Django models should implement a __unicode__ method for string representation. "
                        "A parent class of this model does, but ideally all models should be explicit.")
}


def _is_meta_with_abstract(node):
    if isinstance(node, ClassDef) and node.name == 'Meta':
        for meta_child in node.get_children():
            if not isinstance(meta_child, Assign):
                continue
            if not meta_child.targets[0].name == 'abstract':
                continue
            if not isinstance(meta_child.value, Const):
                continue
                # TODO: handle tuple assignment?
            # eg:
            #    abstract, something_else = True, 1
            if meta_child.value.value:
                # this class is abstract
                return True
    return False


def _has_python_2_unicode_compatible_decorator(node):
    if node.decorators is None:
        return False

    for decorator in node.decorators.nodes:
        if getattr(decorator, 'name', None) == 'python_2_unicode_compatible':
            return True

    return False


def _is_unicode_or_str_in_python_2_compatibility(method):
    if method.name == '__unicode__':
        return True

    if method.name == '__str__' and _has_python_2_unicode_compatible_decorator(method.parent):
        return True

    return False


class ModelChecker(BaseChecker):
    """Django model checker."""
    __implements__ = IAstroidChecker

    name = 'django-model-checker'
    msgs = MESSAGES

    # NOTE: there's a bug in pylint's backwards compatability logic after changing
    # visit method names from visit_class to visit_classdef, see
    # https://bitbucket.org/logilab/pylint/issues/711/new-node-visit-methods-not-backwards
    if PYLINT_VERSION < (1, 5):
        @check_messages('model-missing-unicode')
        def visit_class(self, node):
            return self._visit_classdef(node)
    else:
        @check_messages('model-missing-unicode')
        def visit_classdef(self, node):
            return self._visit_classdef(node)

    def _visit_classdef(self, node):
        """Class visitor."""
        if not node_is_subclass(node, 'django.db.models.base.Model', '.Model'):
            # we only care about models
            return

        for child in node.get_children():
            if _is_meta_with_abstract(child):
                return

            if isinstance(child, Assign):
                grandchildren = list(child.get_children())

                if not isinstance(grandchildren[0], AssignName):
                    continue

                name = grandchildren[0].name
                if name != '__unicode__':
                    continue

                grandchild = grandchildren[1]
                assigned = grandchild.inferred()[0]

                if assigned.callable():
                    return

                self.add_message('E%s01' % BASE_ID, args=node.name, node=node)
                return

            if isinstance(child, FunctionDef) and child.name == '__unicode__':
                if PY3:
                    self.add_message('W%s02' % BASE_ID, args=node.name, node=node)
                return

        # if we get here, then we have no __unicode__ method directly on the class itself

        # a different warning is emitted if a parent declares __unicode__
        for method in node.methods():
            if method.parent != node and _is_unicode_or_str_in_python_2_compatibility(method):
                # this happens if a parent declares the unicode method but
                # this node does not
                self.add_message('W%s03' % BASE_ID, args=node.name, node=node)
                return

        # if the Django compatibility decorator is used then we don't emit a warning
        # see https://github.com/PyCQA/pylint-django/issues/10
        if _has_python_2_unicode_compatible_decorator(node):
            return

        if PY3:
            return

        self.add_message('W%s01' % BASE_ID, args=node.name, node=node)



# Copyright (c) 2018 Alexander Todorov <atodorov@MrSenko.com>

# Licensed under the GPL 2.0: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# For details: https://github.com/PyCQA/pylint-django/blob/master/LICENSE
"""
Various suggestions about JSON http responses
"""

import astroid

from pylint import interfaces
from pylint import checkers
from pylint.checkers import utils
from pylint_django.__pkginfo__ import BASE_ID


class JsonResponseChecker(checkers.BaseChecker):
    """
    Looks for some common patterns when returning http responses containing
    JSON data!
    """

    __implements__ = (interfaces.IAstroidChecker,)

    # configuration section name
    name = 'json-response-checker'
    msgs = {'R%s01' % BASE_ID: ("Instead of HttpResponse(json.dumps(data)) use JsonResponse(data)",
                                'http-response-with-json-dumps',
                                'Used when json.dumps() is used as an argument to HttpResponse().'),
            'R%s02' % BASE_ID: ("Instead of HttpResponse(content_type='application/json') use JsonResponse()",
                                'http-response-with-content-type-json',
                                'Used when HttpResponse() is returning application/json.'),
            'R%s03' % BASE_ID: ("Redundant content_type parameter for JsonResponse()",
                                'redundant-content-type-for-json-response',
                                'Used when JsonResponse() contains content_type parameter. '
                                'This is either redundant or the content_type is not JSON '
                                'which is probably an error.')}

    @utils.check_messages('http-response-with-json-dumps',
                          'http-response-with-content-type-json',
                          'redundant-content-type-for-json-response')
    def visit_call(self, node):
        if (node.func.as_string().endswith('HttpResponse') and node.args
                and isinstance(node.args[0], astroid.Call)
                and node.args[0].func.as_string() == 'json.dumps'):
            self.add_message('http-response-with-json-dumps', node=node)

        if node.func.as_string().endswith('HttpResponse') and node.keywords:
            for keyword in node.keywords:
                if (keyword.arg == 'content_type'
                        and keyword.value.as_string().lower().find('application/json') > -1):
                    self.add_message('http-response-with-content-type-json', node=node)
                    break

        if node.func.as_string().endswith('JsonResponse') and node.keywords:
            for keyword in node.keywords:
                if keyword.arg == 'content_type':
                    self.add_message('redundant-content-type-for-json-response', node=node)
                    break



"""
Ensures that under PY3 django models with a __unicode__ method are flagged
"""
#  pylint: disable=missing-docstring

from django.db import models


class SomeModel(models.Model):  # [model-has-unicode]
    something = models.CharField(max_length=255)
    # no __str__ method

    something.something_else = 1

    def lala(self):
        pass

    def __unicode__(self):
        return self.something



"""
Test to validate that pylint_django doesn't produce
Instance of 'SubFactory' has no 'pk' member (no-member) warnings
"""
# pylint: disable=attribute-defined-outside-init, missing-docstring, too-few-public-methods
import factory

from django import test
from django.db import models


class Author(models.Model):
    name = models.CharField()


class Book(models.Model):
    title = models.CharField()
    author = models.ForeignKey(Author, on_delete=models.CASCADE)


class AuthorFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = 'Author'

    name = factory.Sequence(lambda n: 'Author %d' % n)


class BookFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = 'Book'

    title = factory.Sequence(lambda n: 'Book %d' % n)
    author = factory.SubFactory(AuthorFactory)


class BookTestCase(test.LiveServerTestCase):
    serialized_rollback = True

    def _fixture_setup(self):
        super(BookTestCase, self)._fixture_setup()
        self.book = BookFactory()

    def test_author_is_not_none(self):
        self.assertGreater(self.book.pk, 0)
        self.assertGreater(self.book.author.pk, 0)

        self.assertIsNotNone(self.book.title)
        self.assertIsNotNone(self.book.author.name)



"""
Checks that Pylint does not complain about django FormViews
having too many ancestors
"""
#  pylint: disable=missing-docstring
from django.views.generic import FormView


class SomeView(FormView):
    pass



"""
Checks that Pylint does not complain about attributes and methods
when using Class-based Views
"""
#  pylint: disable=missing-docstring

from django.views.generic import TemplateView


class BoringView(TemplateView):
    # ensure that args, kwargs and request are not thrown up as errors
    def get_context_data(self, **kwargs):
        return {
            'request': self.request,
            'args': self.args,
            'kwargs': self.kwargs
        }



"""
Initial migration which should not raise any pylint warnings.
"""
# pylint: disable=missing-docstring, invalid-name
from django.db import migrations, models


class Migration(migrations.Migration):
    operations = [
        migrations.CreateModel(
            name='TestRun',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True)),
                ('summary', models.TextField()),
                ('environment_id', models.IntegerField(default=0)),
                ('auto_update_run_status', models.BooleanField(default=False)),
            ],
        ),
    ]



"""
Checks that Pylint does not complain about django lazy proxy
when using ugettext_lazy
"""
from django.utils.translation import ugettext_lazy

ugettext_lazy('{something}').format(something='lala')



from django.db.models.fields import files as django_fields


class FileField(django_fields.FieldFile, django_fields.FileField):
    def __init__(self, verbose_name=None, name=None, upload_to='', storage=None, **kwargs):
        django_fields.FileField.__init__(verbose_name, name, upload_to, storage, **kwargs)


class ImageField(django_fields.ImageFieldFile, django_fields.ImageField):
    def __init__(self, verbose_name=None, name=None, width_field=None, height_field=None, **kwargs):
        django_fields.ImageField.__init__(verbose_name, name, width_field, height_field, **kwargs)



"""
Checks that Pylint does not complain about various
methods on Django model fields.
"""
#  pylint: disable=missing-docstring
from __future__ import print_function
from datetime import datetime, date
from decimal import Decimal
from django.db import models


class LotsOfFieldsModel(models.Model):

    bigintegerfield = models.BigIntegerField()
    booleanfield = models.BooleanField(default=True)
    charfield = models.CharField(max_length=40, null=True)
    commaseparatedintegerfield = models.CommaSeparatedIntegerField()
    datetimefield = models.DateTimeField(auto_now_add=True)
    datefield = models.DateField(auto_now_add=True)
    decimalfield = models.DecimalField(max_digits=5, decimal_places=2)
    durationfield = models.DurationField()
    emailfield = models.EmailField()
    filefield = models.FileField(name='test_file', upload_to='test')
    filepathfield = models.FilePathField()
    floatfield = models.FloatField()
    genericipaddressfield = models.GenericIPAddressField()
    imagefield = models.ImageField(name='test_image', upload_to='test')
    ipaddressfield = models.IPAddressField()
    intfield = models.IntegerField(null=True)
    nullbooleanfield = models.NullBooleanField()
    positiveintegerfield = models.PositiveIntegerField()
    positivesmallintegerfield = models.PositiveSmallIntegerField()
    slugfield = models.SlugField()
    smallintegerfield = models.SmallIntegerField()
    textfield = models.TextField()
    timefield = models.TimeField()
    urlfield = models.URLField()

    def boolean_field_tests(self):
        print(self.booleanfield | True)
        print(self.nullbooleanfield | True)

    def string_field_tests(self):
        print(self.charfield.strip())
        print(self.charfield.upper())
        print(self.charfield.replace('x', 'y'))

        print(self.filepathfield.strip())
        print(self.filepathfield.upper())
        print(self.filepathfield.replace('x', 'y'))

        print(self.emailfield.strip())
        print(self.emailfield.upper())
        print(self.emailfield.replace('x', 'y'))

        print(self.textfield.strip())
        print(self.textfield.upper())
        print(self.textfield.replace('x', 'y'))

    def datetimefield_tests(self):
        now = datetime.now()
        print(now - self.datetimefield)
        print(self.datetimefield.ctime())

    def datefield_tests(self):
        now = date.today()
        print(now - self.datefield)
        print(self.datefield.isoformat())

    def decimalfield_tests(self):
        print(self.decimalfield.compare(Decimal('1.4')))

    def durationfield_tests(self):
        now = datetime.now()
        print(now - self.durationfield)
        print(self.durationfield.total_seconds())

    def filefield_tests(self):
        self.filefield.save('/dev/null', 'TEST')
        print(self.filefield.file)
        self.imagefield.save('/dev/null', 'TEST')
        print(self.imagefield.file)

    def numberfield_tests(self):
        print(self.intfield + 5)
        print(self.bigintegerfield + 4)
        print(self.smallintegerfield + 3)
        print(self.positiveintegerfield + 2)
        print(self.positivesmallintegerfield + 1)



"""
Checks that Pylint does not complain about django Forms
"""
#  pylint: disable=missing-docstring,wrong-import-position

from django import forms


class TestForm(forms.Form):
    class Meta:
        pass

    some_field = forms.CharField()

    def clean(self):
        print(self.cleaned_data)
        print(self.fields)
        print(self.error_class)


class TestModelForm(forms.ModelForm):
    class Meta:
        pass


class TestFormSubclass(forms.Form):
    class Meta:
        pass


class TestModelFormSubclass(forms.ModelForm):
    class Meta:
        pass



from astroid import MANAGER, nodes, InferenceError, inference_tip, UseInferenceDefault
from astroid.nodes import ClassDef, Attribute

from pylint_django.utils import node_is_subclass


def is_foreignkey_in_class(node):
    # is this of the form  field = models.ForeignKey
    if not isinstance(node.parent, nodes.Assign):
        return False
    if not isinstance(node.parent.parent, ClassDef):
        return False

    if isinstance(node.func, Attribute):
        attr = node.func.attrname
    elif isinstance(node.func, nodes.Name):
        attr = node.func.name
    else:
        return False
    return attr in ('OneToOneField', 'ForeignKey')


def infer_key_classes(node, context=None):
    for arg in node.args:
        # typically the class of the foreign key will
        # be the first argument, so we'll go from left to right
        if isinstance(arg, (nodes.Name, nodes.Getattr)):
            try:
                key_cls = None
                for inferred in arg.infer(context=context):
                    key_cls = inferred
                    break
            except InferenceError:
                continue
            else:
                if key_cls is not None:
                    break
        elif isinstance(arg, nodes.Const):
            try:
                model_name = arg.value.split('.')[-1]  # can be 'Model' or 'app.Model'
            except AttributeError:
                break

            for module in MANAGER.astroid_cache.values():
                if model_name in module.locals:
                    class_defs = [
                        module_node for module_node in module.lookup(model_name)[1]
                        if isinstance(module_node, nodes.ClassDef)
                        and node_is_subclass(module_node, 'django.db.models.base.Model')
                    ]
                    if class_defs:
                        return iter([class_defs[0].instantiate_class()])
    else:
        raise UseInferenceDefault
    return iter([key_cls.instantiate_class()])


def add_transform(manager):
    manager.register_transform(nodes.CallFunc, inference_tip(infer_key_classes),
                               is_foreignkey_in_class)



from __future__ import absolute_import
from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint_django.__pkginfo__ import BASE_ID


class DjangoInstalledChecker(BaseChecker):
    name = 'django-installed-checker'

    msgs = {
        'F%s01' % BASE_ID: ("Django is not available on the PYTHONPATH",
                            'django-not-available',
                            "Django could not be imported by the pylint-django plugin, so most Django related "
                            "improvements to pylint will fail."),

        'W%s99' % BASE_ID: ('Placeholder message to prevent disabling of checker',
                            'django-not-available-placeholder',
                            'PyLint does not recognise checkers as being enabled unless they have at least'
                            ' one message which is not fatal...')
    }

    @check_messages('django-not-available')
    def close(self):
        try:
            __import__('django')
        except ImportError:
            self.add_message('F%s01' % BASE_ID)



"""
Checks that Pylint does not complain about foreign key sets on models
"""
#  pylint: disable=missing-docstring

from django.db import models


class SomeModel(models.Model):
    name = models.CharField(max_length=20)
    timestamp = models.DateTimeField()


class OtherModel(models.Model):
    something = models.ForeignKey(SomeModel, on_delete=models.CASCADE)
    elsething = models.OneToOneField(SomeModel, on_delete=models.CASCADE)

    def something_doer(self):
        part_a = '%s - %s' % (self.something.name, self.something.timestamp)
        part_b = self.elsething.name
        return part_a, part_b



"""
Ensures that no '__unicode__ missing' warning is emitted if the
Django python3/2 compatability dectorator is used

See https://github.com/PyCQA/pylint-django/issues/10
"""
#  pylint: disable=missing-docstring
from django.utils.encoding import python_2_unicode_compatible
from django.db import models


@python_2_unicode_compatible
class ModelName(models.Model):
    name = models.CharField(max_length=200)

    def __str__(self):
        return self.name
